import logging
from .vin_decoder import *
from .remote_client import DatabricksSQLClient

class QVin(VinDecoder):
    _FEATURES = ['wmi','vds','vis','body_type', 'fuel_type', 'truckDB_Engine_power_bhp', 'weight']
    def __init__(self, credentials=None): 
        if credentials: 
            from databricks import sql
            logging.getLogger("databricks.sql").setLevel(logging.ERROR)
            connector = DatabricksSQLClient(**credentials).execute_statement
        else: connector = spark.sql
        super(QVin, self).__init__(connector)
        self._db_features = QHiveConnector("qirevin.truck_car", connector)

    def lookup_vin(self, vin, query=None, hierarchical=False, feature=False): 
        response=super().lookup_vin(vin, query=query, hierarchical=hierarchical).replace(np.nan, None)
        return self._feature_selection(self._expand_features(response)) if feature else response

    def _feature_selection(self, df):
        return df[self._FEATURES].rename(columns={"truckDB_Engine_power_bhp": "engine_power_bhp"})
    
    def get_distinct(self): return {"body_type": self._distinct(self._db_connector, 'body_type'), "fuel_type": self._distinct(self._db_connector, 'fuel_type')}
    
    def _expand_features(self, df):
        features=self._spark2pandas(self._db_features.selectObj("*"))

        def fuzzy_lookup(x, features, threshold=80):
            match_columns = {"truckDB_Model":"model", "truckDB_Serie":"serie"}
            
            def iterator(cols, x_row, f_table, fuzzy_threshold=None):
                if fuzzy_threshold is None:
                    for fcol,xcol in cols.items():
                        res = f_table[(f_table[fcol].str.lower() == str(x[xcol]).lower())]
                        if not res.empty: return res.iloc[0].squeeze()
                        else: continue
                else:
                    for idx, row in f_table.iterrows():
                        similarity_scores = [
                            fuzz.partial_ratio(str(x_row[xcol]).lower(), str(row[fcol]).lower())
                            for fcol,xcol in cols.items()
                        ]
                        if any(score >= fuzzy_threshold for score in similarity_scores):
                            return row.squeeze()
                return None
            

            condition = (features["truckDB_Make"].str.lower() == str(x['maker']).lower())
            if features[condition].empty: return pd.Series([None] * len(features.columns), index=list(features.columns))
            res=iterator(match_columns, x, features[condition])
            if res is None: res=iterator(match_columns, x, features[condition], 0.9)
            
            return res if res is not None else pd.Series([None] * len(features.columns), index=list(features.columns))

        lookup_result = df.apply(lambda x: fuzzy_lookup(x, features), axis=1)

        return pd.concat([df, lookup_result], axis=1)


from databricks import sql

class DatabricksSQLClient:
    def __init__(self, server_hostname: str, http_path: str, access_token: str):
        self._server_hostname = server_hostname
        self._http_path = http_path
        self._token = access_token

    def execute_statement(self, statement: str):
        with sql.connect(server_hostname=self._server_hostname,
                         http_path=self._http_path,
                         access_token=self._token) as connection:

            with connection.cursor() as cursor:
                cursor.execute(statement)
                return cursor.fetchall_arrow()


import pandas as pd


class QDBConnector:
    def selectObj(self, columns, condition): pass
    def deleteObj(self, idx): pass
    def insertObj(self, new_data): pass
    def updateObj(self, new_data, index=None): pass

class QPDConnector(QDBConnector):
    def __init__(self, table):
        self.table = table
    
    def selectObj(self, columns, condition):
        if isinstance(condition, list): condition = " & ".join(condition)
        return self.table.query(condition)[self.table.columns if columns=="*" else columns].copy()
    
    def deleteObj(self, idx):
        self.table.drop(idx, inplace=True)
    
    def insertObj(self, new_data):
        new_data = pd.DataFrame([new_data])
        self.table = pd.concat([self.table, new_data], ignore_index=True)
    
    def updateObj(self, new_data, index=None):
        if isinstance(new_data, dict): new_data = pd.DataFrame(dict, index=index)
        self.table.update(new_data)
        
    def columns(self):
        return self.table.columns

class QHiveConnector(QDBConnector):
    def __init__(self, table, connector, to_pandas=True):
        self._table = table
        self._connector=connector
        self._columns = self._get_columns(self._connector(f"select * from  {self._table} limit 1"))
        
    def _get_columns(self, row):
        if type(row).__name__ == 'Table': row=row.to_pandas()
        return row.columns
    
    def selectObj(self, columns="*", condition=None):
        print(f"select {columns} from {self._table} {'where '+condition if condition else ''}")
        return self._connector(f"select {columns} from {self._table} {'where '+condition if condition else ''}")
    
    def deleteObj(self, condition):
        print(f"delete from {self._table} where {condition}")
        self._connector(f"delete from {self._table} where {condition}")
    
    def insertObj(self, new_data):
        print(f"insert into {self._table} ({','.join(new_data.keys())}) VALUES ({','.join([qstr(x) for x in new_data.values()])})")
        self._connector(f"insert into table {self._table} ({','.join(new_data.keys())}) VALUES ({','.join([qstr(x) for x in new_data.values()])})")

    def executeQuery(self, query):
        return self._connector(query)
    
    @property
    def columns(self):
        return self._columns
    
    @property
    def table(self):
        return self._table


from .db_client import *
from .external_api import *


def insertValsToHive(data):
    vals=[]
    for k,v in data.items():
        if k in ('other','year'): continue
        vals.append(f"\'{v}\'") if v else vals.append(f"NULL")
    return vals

class VinDecoder:
    def __init__(self, connector):
        self._db_connector =  QHiveConnector("qirevin.out_db", connector)
        self._db_backup_connector = QHiveConnector("qirevin.backup_out_db", connector)
    
        self._vin_parser = VinInfo()

        self._vin_external_api=[VinNHTSA()]

    def _spark2pandas(self, df):
        return df.to_pandas() if type(df).__name__ == 'Table' else df.toPandas()
        
    def lookup_vin(self, vin, query=None, hierarchical=False):
        if all([vin,query]) or not any([vin,query]): Exception('ValueError. Use either vin or query')
        if query: return self._spark2pandas(self._db_connector.selectObj("*", query))
        if isinstance(vin, str): vin=[vin]
        result=[]
        for v in vin:
            try:
                decoded_vin = self._parse_vin(v,hierarchical)
                if decoded_vin is not None: result.append(decoded_vin)
            except Exception as e: 
                print(e)
                continue
        if len(result)==0: return None
        elif len(result)==1: return result[0]
        return pd.concat(result)

    def _parse_vin(self, v,hierarchical):
        vin_info=self._vin_parser.lookup_vin(v)
        if not vin_info: 
            print(f"Unknown {v}")
            return None
        conditions=f"vds = '{vin_info['vds']}' and wmi = '{vin_info['wmi']}' and vis = '{vin_info['vis']}'"
        if hierarchical: conditions=conditions.replace("and", "or")
        result_vin = self._spark2pandas(self._db_connector.selectObj("*", conditions)) # exact match
        if not result_vin.empty: return result_vin
        
        external_api_res = {}
        for api in self._vin_external_api:
            try: external_api_res.update(api.lookup_vin(v))
            except: continue
        insert_data={**vin_info, **external_api_res, "other": {**vin_info.get("other", {}), **external_api_res.get("other", {})}} if external_api_res else vin_info
        self._insert({k:v for k,v in insert_data.items() if v is not None}, self._db_connector)
        return self._spark2pandas(self._db_connector.selectObj("*", conditions))
    
    def export_excel(self, vin, query=None, hierarchical=False, fname="output.xlsx"):
        response = self.lookup_vin(vin=vin, query=query, hierarchical=hierarchical)
        if isinstance(response, pd.DataFrame): 
            response['date_created'] = response['date_created'].dt.tz_localize(None)
            response.to_excel(fname)
        else: Exception('ValueError. No data to write')
            
    def _create_snapshot(self, row):
        if all([row['vds'],row['wmi'],row['vis']]):
            conditions=f"vds = '{row['vds']}' and wmi = '{row['wmi']}' and vis = '{row['vis']}'"
            result_vin = self._spark2pandas(self._db_connector.selectObj("*", conditions))
            if not result_vin.empty:
                insert_data=result_vin.replace(np.nan, None).to_dict("records")[0]
                self._insert_backup({k:v for k,v in insert_data.items() if v is not None},self._db_backup_connector,"date_added_backup")
                self._db_connector.deleteObj(conditions)
        insert_data=row.to_dict()
        insert_data.pop("date_created", None)
        self._insert({k:v for k,v in insert_data.items() if v is not None}, self._db_connector)

    def _distinct(self, table, columns):
        if isinstance(columns, str): columns=[columns]
        q = f"select distinct {','.join(columns)} from {table.table}"
        return self._spark2pandas(self._db_connector.executeQuery(q)).values[:, 0]
        
    def update_db(self, data):
        if isinstance(data, str): data = pd.read_csv(data, index_col=0, low_memory=False)
        elif isinstance(data, list): data = pd.DataFrame(data)
        elif isinstance(data, dict): data = pd.DataFrame.from_dict(data)
        data.apply(lambda x: self._create_snapshot(x), axis=1)

    def _insert(self, data,table):
        data={**{k:None for k in table.columns if k != "date_created"}, **data}
        vals=','.join(insertValsToHive(data))
        q=f"""insert into table {table.table} ({','.join([x for x in data.keys() if x not in ('other', 'year', 'date_created')])},other,year,date_created) VALUES ({vals}, '{json.dumps(data.get('other', {}))}', {data.get("year", None)},current_timestamp())"""
        print(q)
        self._db_connector.executeQuery(q)

    def _insert_backup(self, data,table,date_col="date_added_backup"):
        data={**{k:None for k in table.columns if k !=date_col}, **data}
        vals=','.join(insertValsToHive(data))
        q=f"""insert into table {table.table} ({','.join([x for x in data.keys() if x not in ('other', 'year')])},other,year,{date_col}) VALUES ({vals}, '{json.dumps(data.get('other', {})) if isinstance(data.get('other', {}), dict) else data.get('other')}', {data.get("year", None)},current_timestamp())"""
        print(q)
        self._db_connector.executeQuery(q)